/*

	This is the gtGesture library from BYONDscape.com. Thanks for your support!


	See demo.dm for an example of usage.


Purpose:
	Allow users to enter "gestures" by holding the left mouse button and moving the mouse
	around the BYOND map.

*/


var/gtGesture_gestureHandler/gtGesture_handler


gtGesture_gestureHandler
	var/gestures = list()


	proc/AddGesture(gName, list/L)
		//Register a gesture template.
		var/gtGesture_gesture/G

		if(istype(gName, /gtGesture_gesture))
			G = gName
		else
			G = new (gName, L)

		gestures += G
		return G


	proc/AddReverseGesture(origName, reverseName)
		//Register a copy of an already-added gesture template, drawn from end-to-start.
		//Gestures and their reverses may be registered under the same name if direction
		//is not a concern for your program.
		var/gtGesture_gesture/G

		for(G in gestures)
			if(G.name == origName)
				AddGesture(G.GetReverse(reverseName))
				return


gtGesture_segment
	var
		x1; x2; y1; y2
		minPct; maxPct


	New(xa, ya, xb, yb)
		x1 = xa
		y1 = ya
		x2 = xb
		y2 = yb


	proc/Dist()
		//Calculate the length of the line segment.
		return gtGesture_GetRealDist(x1, y1, x2, y2)


gtGesture_segmentRoster
	var/list/segments = list()
	var/accumDistance = 0


	proc/AddSegment(x1, y1, x2, y2)
		if((x1 == x2) && (y1 == y2)) return

		var/gtGesture_segment/W = new(x1, y1, x2, y2)
		segments += W


	proc/Finalize()
		//Assign each segment a percentage range for later analysis.
		var/basePct = 0

		if(!length(segments)) return null

		for(var/gtGesture_segment/W in segments)
			accumDistance += W.Dist()

		for(var/gtGesture_segment/W in segments)
			W.minPct = basePct
			W.maxPct = basePct + (W.Dist() * 100 / accumDistance)
			basePct = W.maxPct

		return 1


	proc/Scale(minVal, maxVal)
		//Recalculate the endpoints of all line segments so they fall within a specified range.
		var
			minX = 999999; maxX = -999999; minY = 999999; maxY = -999999

		for(var/gtGesture_segment/W in segments)
			if(W.x1 < minX) minX = W.x1
			if(W.x2 < minX) minX = W.x2
			if(W.x1 > maxX) maxX = W.x1
			if(W.x2 > maxX) maxX = W.x2
			if(W.y1 < minY) minY = W.y1
			if(W.y2 < minY) minY = W.y2
			if(W.y1 > maxY) maxY = W.y1
			if(W.y2 > maxY) maxY = W.y2

		for(var/gtGesture_segment/W in segments)
			W.x1 = gtGesture_Yxcabax(minX, maxX, W.x1, minVal, maxVal)
			W.x2 = gtGesture_Yxcabax(minX, maxX, W.x2, minVal, maxVal)
			W.y1 = gtGesture_Yxcabax(minY, maxY, W.y1, minVal, maxVal)
			W.y2 = gtGesture_Yxcabax(minY, maxY, W.y2, minVal, maxVal)


	proc/GetPoint(myPercent)
		//myPercent represents a distance along the total path of the line segments, with
		//0 representing the first point of the first segment and 100 being the endpoint
		//of the last segment.
		if(myPercent > 100) myPercent = 100
		else if(myPercent < 0) myPercent = 0

		var/gtGesture_segment/W
		for(var/gtGesture_segment/W2 in segments)
			if((W2.maxPct + 0.001) >= myPercent) //fixme: math bug! 100 < 100!
				W = W2
				break

		var/xval = gtGesture_Yxcabax(W.minPct, W.maxPct, myPercent, W.x1, W.x2)
		var/yval = gtGesture_Yxcabax(W.minPct, W.maxPct, myPercent, W.y1, W.y2)

		return list(xval, yval)


gtGesture_gesture
	var/gtGesture_segmentRoster/roster
	var/name


	New(myName, list/L)
		name = myName
		roster = new()
		for(var/j = 1; j <= length(L) - 3; j += 2)
			roster.AddSegment(L[j], L[j + 1], L[j + 2], L[j + 3])
		roster.Finalize()


	proc/GetReverse(myName)
		var/list/L = list()
		var/gtGesture_segment/W

		for(var/j = length(roster.segments); j >= 1; j--)
			W = roster.segments[j]
			L += W.x2
			L += W.y2

		W = roster.segments[1]
		L += W.x1
		L += W.y1
		return new /gtGesture_gesture(myName, L)


	proc/Compare(gtGesture_segmentRoster/R, compares = 21)
		var/totalScore = 0
		var/increment = 100 / (compares - 1)
		var/curPct = -increment

		for(var/i = 1; i <= compares; i++)
			curPct += increment
			var/list/W1 = R.GetPoint(curPct)
			var/list/W2 = roster.GetPoint(curPct)
			totalScore += gtGesture_GetRealDist(W1[1], W1[2], W2[1], W2[2])

		totalScore /= compares

		return totalScore


client
	var/gtGesture_segmentRoster/roster
	var
		gtGesture_lastPointX; gtGesture_lastPointY //; gtGesture_lastDrag

	var
		gtGesture_minRelativeX
		gtGesture_maxRelativeX
		gtGesture_minRelativeY
		gtGesture_maxRelativeY
		gtGesture_enforceInputArea = 0


	proc/gtGesture_LimitInputArea(minX, minY, maxX, maxY)
		var/tmpVar

		if(minX > maxX)
			tmpVar = minX
			minX = maxX
			maxX = tmpVar

		if(minY > maxY)
			tmpVar = minY
			minY = maxY
			maxY = tmpVar

		gtGesture_minRelativeX = minX
		gtGesture_maxRelativeX = maxX
		gtGesture_minRelativeY = minY
		gtGesture_maxRelativeY = maxY
		gtGesture_enforceInputArea = 1


	proc/gtGesture_SetWaypoint(theX, theY)
		var/failed = 0
		if(gtGesture_enforceInputArea)
			if(theX < gtGesture_minRelativeX) failed = 1
			if(theX > gtGesture_maxRelativeX) failed = 1
			if(theY < gtGesture_minRelativeY) failed = 1
			if(theY > gtGesture_maxRelativeY) failed = 1

		if(failed)
			del roster
			return

		gtGesture_lastPointX = theX
		gtGesture_lastPointY = theY


	MouseDown(myObj, atom/myLoc)
		. = ..()
		if(istype(myLoc))
			roster = new()
			gtGesture_SetWaypoint(myLoc.x - virtual_eye:x, myLoc.y - virtual_eye:y)


	MouseDrag(srcObj, overObj, srcLoc, atom/myLoc)
		. = ..()
		if(istype(myLoc))
			if(roster)
				roster.AddSegment(gtGesture_lastPointX, gtGesture_lastPointY, \
					myLoc.x - virtual_eye:x, myLoc.y - virtual_eye:y)
				gtGesture_SetWaypoint(myLoc.x - virtual_eye:x, myLoc.y - virtual_eye:y)

			else
				roster = new()
				gtGesture_SetWaypoint(myLoc.x - virtual_eye:x, myLoc.y - virtual_eye:y)


	MouseUp(myObj, myLoc)
		. = ..()
		gtGesture_AnalyzeGesture()


	MouseDrop(srcObj, overObj, srcLoc, atom/myLoc)
		. = ..()
		gtGesture_AnalyzeGesture()


	proc/gtGesture_AnalyzeGesture()
		if(!roster) return
		if(roster.segments.len < 2) return

		roster.Scale(1, 3)

		if(!roster.Finalize())
			del roster
			return

		var/bestScore = 999999
		var/bestName

		for(var/gtGesture_gesture/G in gtGesture_handler.gestures)
			var/comp = G.Compare(roster)
			if(comp < bestScore)
				bestScore = comp
				bestName = G.name

		gtGesture_HandleGesture(bestName, bestScore)

		del roster


	proc/gtGesture_HandleGesture(bestName, bestScore)
		//Override this to dispatch the gestures entered.


proc/gtGesture_Yxcabax(a, b, c, x, y)
	//Map c in (a, b) to equivalent in (x, y)
	if(a == b) return x
	return (y - x) * (c - a) / (b - a) + x


proc/gtGesture_GetRealDist(x1, y1, x2, y2)
	//Return Pythagorean distance between (x1, y1) and (x2, y2)
	return sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
